from ctypes import (CDLL, pointer, c_char, c_char_p, c_int, c_void_p, c_uint32, c_uint8, c_size_t, c_bool, cast, POINTER, CFUNCTYPE, Structure, py_object, addressof)
from enum import Enum
import asyncio
import numpy as np

from ..tools.helpers import input_metadata
from ..tools.helpers import output_metadata
from ..tools.helpers import convert_from_uint8_to_user_dtype


libtpu = CDLL('/usr/lib/libtpu.2.so')
# libtpu = CDLL('libtpu-experimental/build/libtpu.so')
# libtpu = CDLL('/home/d.baburin/iva_tpu_sdk/libtpu.2/build/libtpu.2.so')

__all__ = [
    'TPUDevice',
    'TPUProgram',
    'TPUTensorBufferObject',
    'TPUInference',
    'kPreprocess',
    'kPostprocess',
    'TPUProcessingMode',
    'TPUTensorBufferType',
    'TPUInferenceStatus',
    'load_inference',
    'get_inference',
    'submit_inference',
]



class TPUDevice(Structure):
    
    _fields_ = []

    def __init__(self, dev_name):
        # self.pointer = libtpu.create_device(c_int(dev_no))
        self.pointer = libtpu.create_device(c_char_p(dev_name))

    def __del__(self):
        return libtpu.destroy_device(self.pointer)

    def is_valid(self):
        return libtpu.is_device_valid(self.pointer)

    def init(self):
        return libtpu.init_device(self.pointer)

    def get_error(self):
        return libtpu.get_device_error_message(self.pointer)

    def info(self):
        return libtpu.get_device_info(self.pointer)

    def load_program(self, program):
        return libtpu.load_program(self.pointer, program.pointer)

    def run_inference(self, inference, mode):
        return libtpu.run_inference(self.pointer, inference.pointer, mode)

    def submit_inference(self, inference, mode, callback_fn, context):
        return libtpu.submit_inference(self.pointer, inference.pointer, mode, callback_fn, context)


class TPUProgram(Structure):

    _fields_ = []

    def __init__(self, program_path):
        self.pointer = libtpu.create_program(c_char_p(program_path))

    def __del__(self):
        return libtpu.destroy_program(self.pointer)

    def is_valid(self):
        return libtpu.is_program_valid(self.pointer)

    def info(self):
        return libtpu.get_program_info(self.pointer)

    def get_error(self):
        return libtpu.get_program_error_message(self.pointer)

    def get_batch_size(self):
        return libtpu.get_batch_size(self.pointer)

    def get_input_count(self):
        return libtpu.get_input_count(self.pointer)

    def get_input_size(self, n, raw):
        return libtpu.get_input_size(self.pointer, c_size_t(n), c_bool(raw))

    def get_output_count(self):
        return libtpu.get_output_count(self.pointer)

    def get_output_size(self, n, raw):
        return libtpu.get_output_size(self.pointer, c_size_t(n), c_bool(raw))


class TPUTensorBufferObject(Structure):

    _fields_ = []

    def __init__(self, prog, Type):
        self.pointer = libtpu.create_tensor_buffer_object(prog.pointer, Type)

    def __del__(self):
        return libtpu.destroy_tensor_buffer_object(self.pointer)

    def process_tensor_buffers(self):
        return libtpu.process_tensor_buffers(self.pointer)

    def get_tensor_buffer_ptr(self, n, raw):
        return libtpu.get_tensor_buffer_ptr(self.pointer, c_size_t(n), c_bool(raw))

    def set_user_tensor_buffer_ptr(self, n, ptr):
        return libtpu.set_user_tensor_buffer_ptr(self.pointer, c_size_t(n), ptr)


class TPUInference(Structure):

    _fields_ = []

    def __init__(self, program):
        self.pointer = libtpu.create_inference(program.pointer)

    def __del__(self):
        return libtpu.destroy_inference(self.pointer)

    def get_program(self):
        return libtpu.get_inference_program(self.pointer)

    def get_inputs(self):
        return libtpu.get_inference_inputs(self.pointer)

    def set_inputs(self, buf_obj):
        return libtpu.set_inference_inputs(self.pointer, buf_obj.pointer)

    def set_outputs(self, buf_obj):
        return libtpu.set_inference_outputs(self.pointer, buf_obj.pointer)

    def get_outputs(self):
        return libtpu.get_inference_outputs(self.pointer)

    def get_status(self):
        return libtpu.get_inference_status(self.pointer)

    def get_error(self):
        return libtpu.get_inference_error_message(self.pointer)

CB_POINTER = CFUNCTYPE(None, POINTER(TPUInference), c_int, c_void_p)


def load_inference(data, program, mode=False):
    in_meta = input_metadata(program)
    out_meta = output_metadata(program)

    if not isinstance(data, dict):
        data_ = {}
        for k, d_ in zip(in_meta, (data, )):
            data_.update({k: d_})
        data = data_
    
    tpu_input = TPUTensorBufferObject(program, TPUTensorBufferType.kInputs.value)
    for i, (k, data) in enumerate(data.items()):
        in_tensor = bytearray(data)
        p_in_tensor = (c_uint8 * program.get_input_size(i, mode)).from_buffer(in_tensor)
        tpu_input.set_user_tensor_buffer_ptr(i, cast(p_in_tensor, POINTER(c_uint8)))
    
    out_ctx = []
    tpu_output = TPUTensorBufferObject(program, TPUTensorBufferType.kOutputs.value)
    for o, k in enumerate(out_meta):
        out_tensor = bytearray(np.empty((program.get_output_size(o, mode), ), dtype=np.uint8))
        p_out_tensor = (c_uint8 * program.get_output_size(o, mode)).from_buffer(out_tensor)
        c_out_ptr = cast(p_out_tensor, POINTER(c_uint8))
        out_ctx.append(c_out_ptr)
        tpu_output.set_user_tensor_buffer_ptr(o, c_out_ptr)
    return tpu_input, tpu_output, tuple(out_ctx)

@CB_POINTER
def inference_cb_(inference, result, context):
    cast(context, POINTER(py_object)).contents.value.set()


@CB_POINTER
def inference_cb(inference, result, context):
    cast(context, POINTER(py_object)).contents.value = True


async def submit_inference_(device, program, inference, output_buffer, mode):
    finish_flag = asyncio.Event()
    finish_flag_p = cast(pointer(py_object(finish_flag)), c_void_p)
    # print(f'Flag status before: {finish_flag.is_set()}')
    status = device.submit_inference(inference, mode, inference_cb, finish_flag_p)
    await POINTER(py_object).from_address(addressof(finish_flag_p)).contents.value.wait()
    
    # print(f'Flag status after: {POINTER(py_object).from_address(addressof(finish_flag_p)).contents.value.is_set()}')
    assert status == 0


async def submit_inference(device, program, inference, output_buffer, mode):
    finish_flag = False
    finish_flag_p = cast(pointer(py_object(finish_flag)), c_void_p)
    # print(f'Flag status before: {finish_flag}')
    status = device.submit_inference(inference, mode, inference_cb, finish_flag_p)
    while not POINTER(py_object).from_address(addressof(finish_flag_p)).contents.value:
        await asyncio.sleep(0)
    
    # print(f'Flag status after: {POINTER(py_object).from_address(addressof(finish_flag_p)).contents.value}')
    assert status == 0, 'Inference did not start!'

def get_inference(program, output_buffer, out_ctx, as_dict=True, mode=False):
    out_meta = output_metadata(program)
    collected_data = dict() if as_dict else list()
    
    for i, k in enumerate(out_meta):
        data = np.ctypeslib.as_array(out_ctx[i], shape = (program.get_output_size(i, mode), ))
        data = convert_from_uint8_to_user_dtype(data, out_meta[k]['user_dtype'])
        data = data.reshape(out_meta[k]['user_shape'])
        if as_dict:
            collected_data.update({k: data})
        else:
            collected_data.append(data)
    return collected_data

class TPUProcessingMode(Enum):
    kRaw = 0
    kPreprocess = 1
    kPostprocess = 2
    kFull = 3

class TPUTensorBufferType(Enum):
    kInputs = 0
    kOutputs = 1

class TPUInferenceStatus(Enum):
    kOk = 0
    kError = 1
    kDevError = 2


# TPUDevice
libtpu.create_device.restype = POINTER(TPUDevice)
# libtpu.create_device.argtypes = [c_int]
libtpu.create_device.argtypes = [c_char_p]

libtpu.destroy_device.restype = c_void_p
libtpu.destroy_device.argtypes = [POINTER(TPUDevice)]

libtpu.is_device_valid.restype = c_int
libtpu.is_device_valid.argtypes = [POINTER(TPUDevice)]

libtpu.init_device.restype = c_int
libtpu.init_device.argtypes = [POINTER(TPUDevice)]

libtpu.get_device_error_message.restype = c_char_p
libtpu.get_device_error_message.argtypes = [POINTER(TPUDevice)]

libtpu.get_device_info.restype = c_char_p
libtpu.get_device_info.argtypes = [POINTER(TPUDevice)]

libtpu.load_program.restype = c_int
libtpu.load_program.argtypes = [POINTER(TPUDevice), POINTER(TPUProgram)]

libtpu.run_inference.restype = c_int
libtpu.run_inference.argtypes = [POINTER(TPUDevice), POINTER(TPUInference), c_int]

libtpu.submit_inference.restype = c_int
libtpu.submit_inference.argtypes = [POINTER(TPUDevice), POINTER(TPUInference), c_int, CB_POINTER, c_void_p]

# TPUProgram
libtpu.create_program.restype = POINTER(TPUProgram)
libtpu.create_program.argtypes = [POINTER(c_char)]

libtpu.destroy_program.restype = c_void_p
libtpu.destroy_program.argtypes = [POINTER(TPUProgram)]

libtpu.is_program_valid.restype = c_int
libtpu.is_program_valid.argtypes = [POINTER(TPUProgram)]

libtpu.get_program_info.restype = c_char_p
libtpu.get_program_info.argtypes = [POINTER(TPUProgram)]

libtpu.get_program_error_message.restype = c_char_p
libtpu.get_program_error_message.argtypes = [POINTER(TPUProgram)]

libtpu.get_batch_size.restype = c_size_t
libtpu.get_batch_size.argtypes = [POINTER(TPUProgram)]

libtpu.get_input_count.restype = c_size_t
libtpu.get_input_count.argtypes = [POINTER(TPUProgram)]

libtpu.get_input_size.restype = c_size_t
libtpu.get_input_size.argtypes = [POINTER(TPUProgram), c_size_t, c_bool]

libtpu.get_output_count.restype = c_size_t
libtpu.get_output_count.argtypes = [POINTER(TPUProgram)]

libtpu.get_output_size.restype = c_size_t
libtpu.get_output_size.argtypes = [POINTER(TPUProgram), c_size_t, c_bool]

# TPUTensorBufferObject
libtpu.create_tensor_buffer_object.restype = POINTER(TPUTensorBufferObject)
libtpu.create_tensor_buffer_object.argtypes = [POINTER(TPUProgram), c_int]

libtpu.destroy_tensor_buffer_object.restype = c_void_p
libtpu.destroy_tensor_buffer_object.argtypes = [POINTER(TPUTensorBufferObject)]

libtpu.process_tensor_buffers.restype = c_void_p
libtpu.process_tensor_buffers.argtypes = [POINTER(TPUTensorBufferObject)]

libtpu.get_tensor_buffer_ptr.restype = POINTER(c_uint8)
libtpu.get_tensor_buffer_ptr.argtypes = [POINTER(TPUTensorBufferObject), c_size_t, c_bool]

libtpu.set_user_tensor_buffer_ptr.restype = POINTER(c_uint8)
libtpu.set_user_tensor_buffer_ptr.argtypes = [POINTER(TPUTensorBufferObject), c_size_t, POINTER(c_uint8)]

# TPUInference
libtpu.create_inference.restype = POINTER(TPUInference)
libtpu.create_inference.argtypes = [POINTER(TPUProgram)]

libtpu.destroy_inference.restype = c_void_p
libtpu.destroy_inference.argtypes = [POINTER(TPUInference)]

libtpu.get_inference_program.restype = POINTER(TPUProgram)
libtpu.get_inference_program.argtypes = [POINTER(TPUInference)]

libtpu.get_inference_inputs.restype = POINTER(TPUTensorBufferObject)
libtpu.get_inference_inputs.argtypes = [POINTER(TPUInference)]

libtpu.set_inference_inputs.restype = POINTER(TPUTensorBufferObject)
libtpu.set_inference_inputs.argtypes = [POINTER(TPUInference), POINTER(TPUTensorBufferObject)]

libtpu.get_inference_outputs.restype = POINTER(TPUTensorBufferObject)
libtpu.get_inference_outputs.argtypes = [POINTER(TPUInference)]

libtpu.set_inference_outputs.restype = POINTER(TPUTensorBufferObject)
libtpu.set_inference_outputs.argtypes = [POINTER(TPUInference), POINTER(TPUTensorBufferObject)]

libtpu.get_inference_status.restype = c_int
libtpu.get_inference_status.argtypes = [POINTER(TPUInference)]

libtpu.get_inference_error_message.restype = c_char_p
libtpu.get_inference_error_message.argtypes = [POINTER(TPUInference)]



# pylint: disable=E0611,C0103,R0914,W1514,R1735,R1734,C0206

import asyncio
from typing import Dict, List
from time import time
from zipfile import ZipFile
import tempfile
import json
import os
import shutil
import numpy as np

from ..pytpu import TPUDevice           # type: ignore
from ..pytpu import TPUProgram          # type: ignore
from ..pytpu import TPUInference        # type: ignore
from ..pytpu import TPUProcessingMode   # type: ignore
from ..pytpu import load_inference      # type: ignore
from ..pytpu import get_inference       # type: ignore
from ..pytpu import submit_inference
from ..tools.helpers import to_raw
from ..tools.helpers import STR_TO_DTYPE
from ..tools.helpers import get_tpu_devices

__all__ = [
    'get_fps',
]


async def run_inference(name: str, device: TPUDevice, program: TPUProgram, inference: TPUInference,
                        data_list: List[Dict[str, np.ndarray]]) -> None:

    mode = TPUProcessingMode.kFull.value

    if not bool(program.is_valid()):
        print('\n\nProgram Error: {}\n'.format(program.get_error().decode('utf-8')))

    for ii, data in enumerate(data_list):
        tpu_input, tpu_output, out_ctx = load_inference(data, program)
        inference.set_inputs(tpu_input)
        inference.set_outputs(tpu_output)
        await submit_inference(device, program, inference, tpu_output, mode)
        # data = get_inference(program, tpu_output, out_ctx, as_dict=True, )

        if (ii + 1) % 10 == 0:
            print(f'{name}: finish {ii + 1} iteration')


def get_fps(program_path: str, raw: bool = False, n_queries: int = 100, n_proc: int = 1) -> float:

    print(f'\nStart measure performance for program: {program_path}')
    print(f'\nConfiguration: RAW = {raw}; queries = {n_queries}; processes = {n_proc}')

    tpu_device_names = get_tpu_devices()

    with tempfile.TemporaryDirectory() as tempdir:
        with ZipFile(program_path, 'r') as zip_obj:
            zip_obj.extractall(tempdir)

        with open(os.path.join(tempdir, 'metadata.json'), 'r') as metadata_file:
            metadata = json.load(metadata_file)

        if raw is True:
            with open(os.path.join(tempdir, 'metadata.json'), 'w') as metadata_file:
                metadata = to_raw(metadata)
                json.dump(metadata, metadata_file)

            with tempfile.NamedTemporaryFile() as temp_file:
                program_path = os.path.join(tempdir, 'program_raw.tpu')
                shutil.make_archive(temp_file.name, 'zip', tempdir)
                os.rename(temp_file.name + '.zip', program_path)

            print(f'Raw program saved to {program_path}')

        layers_param = dict()
        for _, region in metadata['inputs'].items():
            for name, inp in region.items():
                layers_param[inp['anchor']] = inp

        data_list = list()
        for _ in range(n_queries):
            for name in layers_param:
                if raw:
                    data_shape = (1, layers_param[name]['size'], )
                    data_type = np.int8
                else:
                    data_shape = layers_param[name]['user_shape']
                    data_type = STR_TO_DTYPE[layers_param[name]['user_dtype']]

                generated_data = (np.random.rand(*data_shape) * 2 - 1) * 100
                generated_data = generated_data.astype(data_type)
                generated_data_dict = {name: generated_data}
                data_list.append(generated_data_dict)

        # input_name = list(layers_param.keys())[0]
        # program_name = os.path.basename(program_path)[:-4]
        # root_dir = os.path.dirname(program_path)
        # with open(os.path.join(root_dir, program_name + '_input.bin'), 'w') as f:
        #     data_list[0][input_name].tofile(f)

        batch = max([layers_param[name]['user_shape'][0] for name in layers_param])

        inference_processes = list()
        device_num = len(tpu_device_names)
        tpu_devices = list()
        tpu_program = TPUProgram(program_path.encode('utf-8'))
        for dev_ii, device_name in enumerate(tpu_device_names):
            tpu_device = TPUDevice(device_name.encode('utf-8'))
            
            if tpu_device.init() < 0:
                raise ConnectionError('Cannot init device')
            
            tpu_device.load_program(tpu_program)
            tpu_devices.append(tpu_device)
            inferences = [TPUInference(tpu_program) for _ in range(n_proc)]
            inference_processes.extend([run_inference(
                name=f'Device {device_name} process {ii}',
                device=tpu_devices[dev_ii],
                program=tpu_program,
                inference=inferences[ii],
                data_list=data_list[ii:len(data_list):n_proc * device_num])
                                        for ii in range(n_proc)])

        start_time = time()
        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.gather(*inference_processes))
        
        total_inference_time = time() - start_time

        fps = n_queries * batch / total_inference_time

        print(f'Estimated FPS = {fps}')

    return fps



from .libtpu_bindings import TPUDevice
from .libtpu_bindings import TPUProgram 
from .libtpu_bindings import TPUTensorBufferObject
from .libtpu_bindings import TPUInference
from .libtpu_bindings import TPUProcessingMode
from .libtpu_bindings import TPUTensorBufferType
from .libtpu_bindings import TPUInferenceStatus
from .libtpu_bindings import load_inference
from .libtpu_bindings import get_inference
from .libtpu_bindings import submit_inference


__all__ = [
    'TPUDevice',
    'TPUProgram',
    'TPUTensorBufferObject',
    'TPUInference',
    'TPUProcessingMode',
    'TPUTensorBufferType',
    'TPUInferenceStatus',
    'load_inference',
    'get_inference',
    'submit_inference',
]



import argparse
import logging
import os
from pytpu.tools.helpers import get_io_from_bin
from pytpu.tools.tpu_runner import TpuRunner
# from pytpu.tools.pyrun_tpu import pyrun_tpu

LOGGER = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(prog='pyrun_tpu',
                                     description='Python command line tool to run tpu programs')

    parser.add_argument('-p', nargs='?', help='Path to the program.tpu', type=str)
    parser.add_argument('-i', nargs='?', help='Inputs binary files path', type=str, default='')
    parser.add_argument('-o', nargs='?', help='Save output binary files path', type=str, default='')
    parser.add_argument('-t', nargs='?', help='Enable execution time calculation', type=str)
    parser.add_argument('-debug', help='Enable DEBUG messages', action='store_true')
    parser.add_argument('-pyproc', action='store_true', help='Enable python pre- and post-processing')

    args = parser.parse_args()

    if args.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO
    logging.basicConfig(level=log_level)
    LOGGER.info(f'Use logging level: {log_level}')

    if args.i is None:
        LOGGER.info('WARNING: no input files are set')
        inputs = []
    else:
        inputs = args.i.split(',')
    input_data, _ = get_io_from_bin(args.p, inputs, 'inputs')

    runner = TpuRunner(program_path=args.p,
                       loop=None,
                       sync=False,
                       pyprocessing=args.pyproc)
    output_data = runner(input_data)
    # output_data = pyrun_tpu(args.p, input_data)

    # Write outputs to binary files
    if args.o is None:
        LOGGER.info('WARNING: no output files are set to store the results')
    else:
        output_files = args.o.split(',')
        _, output_names = get_io_from_bin(args.p, None, 'outputs')
        for idx, name in enumerate(output_names):
            if not os.path.exists(os.path.dirname(output_files[idx])):
                os.makedirs(os.path.dirname(output_files[idx]))
            with open(output_files[idx], 'wb') as io_file:
                io_file.write(output_data[name].tobytes())
                LOGGER.info(f'Writing output {name} to {output_files[idx]}')



# coding: utf-8
"""Setup script for IVA TPU."""
# import os
from setuptools import setup

with open('README.md', 'r') as fh:
    long_description = fh.read()

# with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, 'VERSION'), 'r') as f:
#     VERSION = f.readline()
#     print(f'Install version: {VERSION}')

setup(
    name='pytpu',
    packages=['pytpu', 'pytpu.tools', 'pytpu.pytpu', 'pytpu.scripts'],
    # packages=find_packages(),
    # version=VERSION,
    version="14.6.0",
    author="Alexey Antipin",
    author_email="a.antipin@iva-tech.ru",
    description="TPU Python API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://git.mmp.iva-tech.ru/tpu_sw/iva_tpu_sdk",
    install_requires=[
        'asyncio',
        'numpy',
        'tpu_tlm_is~=0.2.1,>=0.2.1.0',
    ],
    extras_require={
        'test': [
            # Tests
            'pytest',
            'pytest-xdist',
            'pytest-rerunfailures',
            'pytest-cov',
            'Pillow',
            'tpu_tlm[test]~=0.2.0',
            'iva_applications',
            # Code Quality
            'flake8',
            'mypy',
            'pycodestyle',
            'pydocstyle',
            'pylint',
        ],
    },
    zip_safe=False,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'run_get_fps = pytpu.scripts.run_get_fps:main',
            'pyrun_tpu = pytpu.scripts.pyrun_tpu_cli:main'
        ]
    },
)
